use std::{
    io::{Read, Write},
    sync::mpsc::{self, Receiver, Sender, SyncSender},
    thread,
    time::Duration,
};

use anyhow::Context;
use serial::{BaudRate::BaudOther, SerialPort};

use crate::{
    aprs::CommandHandler, config::Config, img::ImgManager, packet::FecPacket, ssdv::ssdv_encode,
};

const IMAGE_PACKET_QUEUE_LENGTH: usize = 1024;

pub struct Controller {
    config: Config,
}

impl Controller {
    pub fn new(config: Config) -> Self {
        Self { config }
    }

    pub fn run_forever(self) {
        let (img_tx, img_rx) = mpsc::sync_channel(IMAGE_PACKET_QUEUE_LENGTH);
        let (cmd_tx, cmd_rx) = mpsc::channel();

        thread::spawn(|| Self::tx_thread(img_rx));
        {
            let config = self.config.clone();
            thread::spawn(|| Self::aprs_thread(config, cmd_tx));
        }

        let mut manager = ImgManager::new(self.config.paths.clone(), cmd_rx);

        loop {
            while let Some((idx, bytes, lossless)) = manager.next() {
                if let Err(e) = self.process_image(&bytes, idx, &img_tx, lossless) {
                    eprintln!("Error processing image: {e}");
                }
            }

            // after we don't find anything, sleep for a bit while
            // we wait for new images
            thread::sleep(Duration::from_secs(1));
        }
    }

    fn process_image(
        &self,
        img: &[u8],
        idx: u8,
        tx: &SyncSender<FecPacket>,
        lossless: bool,
    ) -> anyhow::Result<()> {
        for p in ssdv_encode(&self.config.callsign, img, idx, lossless)? {
            tx.send(p.into()).context("TX thread died")?;
        }

        Ok(())
    }

    // manages incoming APRS packets
    // used to request HD images, as well as
    // to initiate burst/cutdown
    fn aprs_thread(config: Config, tx: Sender<u8>) {
        let mut handler = CommandHandler::new(
            config.callsign,
            config.secret,
            config.burst_command,
            config.cutdown_command,
            tx,
        );

        handler.process_forever();
    }

    // manages our transceiver
    // TODO currently very hacky, because we're going to rip
    // most of this out when we stop using a lobotomized NPR-70
    // and move to just accessing the RF4463 directly over SPI
    fn tx_thread(rx: Receiver<FecPacket>) {
        let mut uart = serial::open("/dev/ttyACM0").unwrap();
        uart.reconfigure(&|settings| settings.set_baud_rate(BaudOther(921600)))
            .unwrap();
        uart.set_timeout(Duration::from_millis(5)).unwrap();

        while let Ok(packet) = rx.recv() {
            uart.write_all(&packet.0).unwrap();
            let mut buf = [0; 1];
            if let Ok(1) = uart.read(&mut buf) {
                if buf[0] == b'F' {
                    // uC is full. Need to wait until we receive
                    // the empty signal
                    while buf[0] != b'E' {
                        let _ = uart.read(&mut buf);
                    }
                }
            }
        }
    }
}
