use image::{imageops::FilterType, io::Reader as ImageReader, ImageOutputFormat};
use std::{fs, io::Cursor, path::PathBuf, sync::mpsc::Receiver};

const IMG_DIM: u32 = 1024;
const JPEG_QUALITY: u8 = 100; // let ssdv do the compression

#[derive(Debug)]
struct ImgInfo {
    path: PathBuf,
    path_idx: usize,
    size_bytes: u64,
    txed: bool,
    found_on_iter: u64,
}

impl ImgInfo {
    pub fn compressed_jpeg_bytes(&self) -> anyhow::Result<Vec<u8>> {
        let img = ImageReader::open(&self.path)?.decode()?;

        // want a 512x512 image
        // first - resize so that the smaller dimension is 512x512
        let img = if img.width() > img.height() {
            img.resize(10_000, IMG_DIM, FilterType::Triangle)
        } else {
            img.resize(IMG_DIM, 10_000, FilterType::Triangle)
        };

        // second - crop to 512x512
        let img = img.crop_imm(
            (img.width() - IMG_DIM) / 2,
            (img.height() - IMG_DIM) / 2,
            IMG_DIM,
            IMG_DIM,
        );

        let mut bytes = vec![];

        let img = img.to_rgb8();
        img.write_to(
            &mut Cursor::new(&mut bytes),
            ImageOutputFormat::Jpeg(JPEG_QUALITY),
        )?;

        Ok(bytes)
    }

    pub fn lossless_jpeg_bytes(&self) -> anyhow::Result<Vec<u8>> {
        let img = ImageReader::open(&self.path)?.decode()?;

        // turn image into a square, making it bigger in the process
        // must be divisible by 16
        let mut dim = img.width().max(img.height());
        let offset = dim % 16;
        dim += (16 - offset) % 16;

        img.crop_imm(0, 0, dim, dim);

        let mut bytes = vec![];

        img.write_to(
            &mut Cursor::new(&mut bytes),
            ImageOutputFormat::Jpeg(JPEG_QUALITY),
        )?;

        Ok(bytes)
    }
}

pub struct ImgManager {
    paths: Vec<PathBuf>,
    imgs: Vec<ImgInfo>,
    // used to receive image requests, which will be sent at full resolution
    rx: Receiver<u8>,

    // state for iterator
    iter: u64,
    path_idx: usize,
    found: bool,
    tx_idx: u8,
    // map between sent image ids, the the images
    // in the array above
    // used for decoding rx requests
    img_map: [usize; 256],
}

impl ImgManager {
    pub fn new(paths: Vec<PathBuf>, rx: Receiver<u8>) -> Self {
        Self {
            paths,
            imgs: vec![],
            rx,

            iter: 0,
            path_idx: 0,
            found: false,
            tx_idx: 0,
            img_map: [0; 256],
        }
    }

    // not making this an iterator since
    // it's entirely valid to have a Some(...)
    // after a None
    pub fn next(&mut self) -> Option<(u8, Vec<u8>, bool)> {
        loop {
            if let Ok(req) = self.rx.try_recv() {
                let img = self
                    .img_map
                    .get(usize::from(req))
                    .and_then(|id| self.imgs.get(*id))
                    .and_then(|img| img.lossless_jpeg_bytes().ok());

                if let Some(img) = img {
                    return Some((self.inc_tx(None), img, true));
                }
            }

            if let Err(e) = self.scan_path() {
                eprintln!(
                    "Error scanning {}: {}",
                    self.paths[self.path_idx].display(),
                    e
                );
            }

            // new images
            let new = self
                .imgs
                .iter_mut()
                .enumerate()
                .filter(|(_, img)| !img.txed) // non-sent images only
                .filter(|(_, img)| img.path_idx == self.path_idx)
                .filter(|(_, img)| img.found_on_iter == self.iter)
                .max_by_key(|(_, img)| img.size_bytes);

            if let Some((new_id, new)) = new {
                new.txed = true;
                let bytes = new.compressed_jpeg_bytes();

                self.inc_iter(true);

                // not perfect. ideally we would keep looking through the new images
                // if this Err'd
                if let Ok(x) = bytes {
                    return Some((self.inc_tx(Some(new_id)), x, false));
                }
            }

            // old images
            let old = self
                .imgs
                .iter_mut()
                .enumerate()
                .filter(|(_, img)| !img.txed) // non-sent images only
                .filter(|(_, img)| img.path_idx == self.path_idx)
                .max_by_key(|(_, img)| img.size_bytes);

            if let Some((old_idx, old)) = old {
                old.txed = true;
                let bytes = old.compressed_jpeg_bytes();

                self.inc_iter(true);

                if let Ok(x) = bytes {
                    return Some((self.inc_tx(Some(old_idx)), x, false));
                }
            }

            // we got all the way through an iteration without finding anything
            // give up for now
            if !self.inc_iter(false) {
                return None;
            }
        }
    }

    fn inc_iter(&mut self, found: bool) -> bool {
        self.path_idx += 1;

        self.found = self.found || found;

        if self.path_idx == self.paths.len() {
            self.path_idx = 0;
            self.iter += 1;

            let found = self.found;
            self.found = false;
            return found;
        }

        true
    }

    fn inc_tx(&mut self, id: Option<usize>) -> u8 {
        let idx = self.tx_idx;
        self.tx_idx = self.tx_idx.wrapping_add(1);

        if let Some(id) = id {
            self.img_map[usize::from(idx)] = id;
        }

        idx
    }

    fn scan_path(&mut self) -> anyhow::Result<()> {
        // imgs we don't already have
        let new_imgs: Vec<_> = fs::read_dir(&self.paths[self.path_idx])?
            // only entries we can actually read without any errors
            .filter_map(|p| {
                p.ok()
                    .and_then(|pp| pp.metadata().ok().map(|meta| (pp.path(), meta)))
            })
            // only files
            .filter(|(p, _)| p.is_file())
            // only new files
            .filter(|(p, _)| !self.imgs.iter().any(|i| i.path == p.as_path()))
            .collect();

        for (path, metadata) in new_imgs {
            let size_bytes = metadata.len();

            self.imgs.push(ImgInfo {
                path,
                path_idx: self.path_idx,
                size_bytes,
                txed: false,
                found_on_iter: self.iter,
            });
        }

        Ok(())
    }
}
